using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class Controller : MonoBehaviour
{
	
	public GameObject player;
	private Camera cam;
	private MouseOrbit mouseorbit;
	private Level level;
	
	public GameObject piecemaskprefab;
	
	private GameObject[] piecemask;
	
	//public GameObject[] selector;
	public float movespeed = 0.7f;
	
	public float lerplength = 1.0f;
	
	public float lerptime = 10.0f;
	
	Vector4 pivot_offset_vector_x;
	Vector4 pivot_offset_vector_y;
	Vector3 pivot_from;
	Vector3 pivot_to;
	Vector3 pivot_axis;
	
	public bool animdone=true;
	
	private const int masklength=16;
	private const int maskcount=2*masklength;
	// Use this for initialization
	void Awake ()
	{
		piecemask = new GameObject[maskcount];
		for (int i=0; i<maskcount; i++)
		{
			GameObject pm = Instantiate(piecemaskprefab) as GameObject;
			pm.transform.parent = gameObject.transform;
			pm.renderer.enabled = false;
			piecemask [i] = pm;
		}
		
		level = GetComponent<Level>();
		cam = Camera.mainCamera;
		mouseorbit = cam.GetComponent<MouseOrbit>();
		//selector[0].renderer.enabled = false;
	}
	
	private bool AtLeastPartiallyVisible (Face f, int camdir)
	{
		Vector3 camforward = - DodecaData.vertices[camdir];
		Vector3[] edgedirs = DodecaData.EdgeDirs [(int)f.side];
		foreach (Vector3 v in edgedirs)
		{
			//half way must be visible - i.e., must give face f
			Vector3 halfway = f.Center() + v * 1.1f / 2.0f;
						
			Ray halfwayray = new Ray(halfway - 100 * camforward, camforward);
			RaycastHit hit;
			if (Physics.Raycast(halfwayray, out hit, 1000, 1 << 10))
			{
				Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
				if (newface.ValueEquals(f))
				{
					return true;
				}
			}
		}
		return false;
	}
	
	//half obscured - no more, no less
	public bool ExactlyPartiallyVisible (Face f, int viewdir)
	{
		bool partiallyvisible = false;
		bool partiallyobscured = false;
		Vector3 camforward = -DodecaData.vertices [viewdir];
		Vector3[] edgedirs = DodecaData.EdgeDirs [(int)f.side];
		foreach (Vector3 v in edgedirs)
		{
			//half way must be visible - i.e., must give face f
			Vector3 halfway = f.Center() + v * 1.1f / 2.0f;
						
			Ray halfwayray = new Ray(halfway - 100 * camforward, camforward);
			RaycastHit hit;
			if (Physics.Raycast(halfwayray, out hit, 1000, 1 << 10))
			{
				Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
				if (newface.ValueEquals(f))
				{
					partiallyvisible = true;
					if (partiallyobscured)
					{
						return true;
					}
				}
				else
				{
					partiallyobscured = true;
					if (partiallyvisible)
					{
						return false;
					}
				}
			}
		}
		return false;
	}
	
	public bool FullyVisible (Face f, out Face obscuringface, int viewdir)
	{
		Vector3 camforward = -DodecaData.vertices[viewdir];
		obscuringface = default(Face);
		Vector3[] edgedirs = DodecaData.EdgeDirs [(int)f.side];
		foreach (Vector3 v in edgedirs)
		{
			//half way must be visible - i.e., must give face f
			Vector3 halfway = f.Center() + v * 1.1f / 2.0f;
						
			Ray halfwayray = new Ray(halfway - 100 * camforward, camforward);
			RaycastHit hit;
			if (Physics.Raycast(halfwayray, out hit, 1000, 1 << 10))
			{
				Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
				if (newface.ValueEquals(f) == false)
				{
					obscuringface = newface;
					return false;
				}
			}
		}
		return true;
	}
	
	//don't need to check that f is already visible - not worthwhile
	public List<Face> GetAllowedFaces (Face f, int viewingdir)
	{
		Vector3 camforward = -DodecaData.vertices [viewingdir];
		List<Face> allowed = new List<Face>();
		Vector3[] edgedirs = DodecaData.EdgeDirs [(int)f.side];		
		
		foreach (Vector3 v in edgedirs)
		{
			//half way must be visible - i.e., must give face f
			Vector3 halfway = f.Center() + v * 1.1f / 2.0f;
			
			bool success = true;
			
			
			Ray halfwayray = new Ray(halfway - 100 * camforward, camforward);
			RaycastHit hit;
			if (Physics.Raycast(halfwayray, out hit, 1000, 1 << 10))
			{
				Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
				if (newface.ValueEquals(f) == false)
				{
					success = false;
				}
			
			}
			else
			{
				success = false;
			}
			
			if (success)
			{
				Vector3 fullway = f.Center() + v * 1.1f;
				Ray fullwayray = new Ray(fullway - 100 * camforward, camforward);
			
				if (Physics.Raycast(fullwayray, out hit, 1000, 1 << 10))
				{
					
					Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
					/*Face newface2 = level.GetFaceFromNormalRay2(hit.point, hit.normal);
					if (newface.ValueEquals(newface2) == false)
					{
						Debug.Log(
							"right " + newface.Print() + "\n"
							+ "wrong " + newface2.Print() + "\n"
							+ "center " + newface.coord.ToVector3() + "\n"
							+ "approx center" + (hit.point - hit.normal * 1.4f) + "\n"
							+ "raycasthit " + hit.point + "\n"
							+ "distance " + Vector3.Distance(hit.point, newface.coord.ToVector3()) + "\n"
							+ "wrongcenter " + newface2.coord.ToVector3() + "\n"
							);
						Debug.DrawLine(hit.point,newface.Center(),Color.red,5);
					}*/
					
					if (IsOverlapping(f, newface, viewingdir) == false)
					{
						allowed.Add(newface);
					}
				
				}

			}
		
		}
		return allowed;
				
	}
	
	//checks that other isn't overlapping main (where we know main to be visible)
	private bool IsOverlapping (Face main, Face other, int viewingdir)
	{
		Vector3 main_center = main.Center();
		Vector3 other_center = other.Center();
		Vector3 forward = -DodecaData.vertices [viewingdir];
		
		//need to check that center is far-enough away on the camera plane
		Vector3 d = other_center - main_center;
		Vector3 other_projected = main_center + d - (Vector3.Project(d, forward));
		float mag = (other_projected - main_center).magnitude;
		return (mag < 1);
							
	}
	
	public void Reset ()
	{
		lerptime = 10;
		animdone = true;
	}	
	
	private bool FollowFingerInput (out Face f)
	{
		bool allowed = false;
		f = default(Face);
		Vector3 viewport_click = cam.ScreenToViewportPoint(Input.mousePosition);
		viewport_click = new Vector3(viewport_click.x, viewport_click.y, 0);
									
		List<Face> allowedfaces = GetAllowedFaces(level.playerface, mouseorbit.targetview);
				
		bool overface = false;
		Face clickedface = default(Face);				
		var ray = Camera.main.ScreenPointToRay(Input.mousePosition);
		RaycastHit hit;
		if (Physics.Raycast(ray, out hit, 1000, 1 << 10))
		{
			overface = true;
			clickedface = level.GetFaceFromNormalRay(hit.point, hit.normal);
		}
				
				
		{
			float min = 1000;
										
			foreach (Face cand_f in allowedfaces)
			{
				//if clicked on allowed face directly, choose it (some blech problems with my code, 
				//this check catches the most obvious crap cases)
				if (overface && clickedface.ValueEquals(cand_f))
				{
					f = cand_f;
					min = -100;
				}
						
				Vector3 candface_screen = cam.WorldToViewportPoint(cand_f.Center());
				candface_screen = new Vector3(candface_screen.x, candface_screen.y, 0);						
						
				float d = Vector3.Distance(viewport_click, candface_screen);
				if (d < min)
				{
					min = d;
					f = cand_f;
				}
			}
								
			//need to check that you're not just clicking on your own square
			if (f.ValueEquals(level.playerface) == false)
			{												
				//need to check that this is moving towards the cursor		
				Vector3 curpos_screen = cam.WorldToViewportPoint(level.playerface.Center());
				curpos_screen = new Vector3(curpos_screen.x, curpos_screen.y);	
						
				float d = Vector3.Distance(viewport_click, curpos_screen);
				if (d > min)
				{
					if (overface && clickedface.ValueEquals(level.playerface))
					{
						//not valid
					}
					else
					{
						allowed = true;
					}
				}
						
						
			}
				
		}	
	
		return allowed;
	}
	
	public bool controlling=false;
	// Update is called once per frame
	void Update ()
	{
	
		if (Editor.editmode || level.leaving)
		{
			return;
		}
		
		if (controlling == false)
		{
			if (InputManager.JustPressedMove())
			{
				controlling = true;
			}
		}
		else
		{
			if (InputManager.JustReleasedMove())
			{
				controlling = false;
			}
		}
		
		if (lerptime > lerplength && controlling)
		{
			//Debug.DrawRay(f.Center(), f.side.Normal(), Color.white);			
			if ((level.Finished() == false) && mouseorbit.state == MouseOrbit.CameraState.iso)
			{				
				
				Face f;
				bool allowed = FollowFingerInput(out f);
				
				if (allowed)
				{	
					animdone = false;
					SnapPlayer();
					Vector3 camforward = cam.transform.forward;
					
					Vector3 going_from = level.playerface.Center();
					Vector3 going_to = f.Center();
					
					//find edge in direction you want
					List<Vector3> source_edgedirs = level.playerface.RealEdgeDirs();
					Vector3 going_to_screenplane = 
							Utility.LinePlaneIntersection(going_to, camforward, going_from, DodecaData.SideNormal [level.playerface.side]);
					
					Vector3 movingfrom_dir = default(Vector3);
					{
						float min = 1000;
						foreach (Vector3 edgedir in source_edgedirs)
						{
							//world space
							Vector3 source_mid = going_from + edgedir;
							//project to camera plane
							Vector3 source_mid_screenplane = 
							Utility.LinePlaneIntersection(source_mid, camforward, going_from, DodecaData.SideNormal [level.playerface.side]);
							float dist = Vector3.Distance(going_to_screenplane, source_mid_screenplane);
							if (dist < min)
							{
								min = dist;
								movingfrom_dir = edgedir;
							}
						}
					}
					
					List<Vector3> target_edgedirs = f.RealEdgeDirs();
					Vector3 going_from_screenplane = 
							Utility.LinePlaneIntersection(going_from, camforward, going_to, DodecaData.SideNormal [f.side]);
					
					Vector3 movingto_dir = default(Vector3);
					{
						float min = 1000;
						foreach (Vector3 edgedir in target_edgedirs)
						{
							//world space
							Vector3 target_mid = going_to + edgedir;
							//project to camera plane
							Vector3 target_mid_screenplane = 
							Utility.LinePlaneIntersection(target_mid, camforward, going_to, DodecaData.SideNormal [f.side]);
							float dist = Vector3.Distance(going_from_screenplane, target_mid_screenplane);
							if (dist < min)
							{
								min = dist;
								movingto_dir = edgedir;
							}
						}
					}
					Vector3 coord_from = level.playerface.coord.ToVector3();
					Vector3 coord_to = f.coord.ToVector3();
					
					pivot_from = going_from + movingfrom_dir;
					pivot_to = going_to + movingto_dir;
					
					if (level.playerface.coord.ValueEquals(f.coord))
					{
						pivot_from = coord_from + (pivot_from - coord_from) * hoverheight;
						pivot_to = coord_to + (pivot_to - coord_to) * hoverheight;
					}
					else//if parallel and/or with different parent
					{
						pivot_from = pivot_from + DodecaData.SideNormal [f.side].normalized * (hoverheight - 1.0f);
						pivot_to = pivot_to + DodecaData.SideNormal [f.side].normalized * (hoverheight - 1.0f);
					}
					
											 
					lerptime = 0;
					
					float diff = Vector3.Angle(DodecaData.SideNormal [f.side], DodecaData.SideNormal [level.playerface.side]);
					if (diff < 40)
					{
						//moving parallel
						lerplength = 1.0f;
					}
					else
					{
						if (f.coord.ValueEquals(level.playerface.coord))
						{
							lerplength = 1.3333333f;
						}
						else
						{
							lerplength = 0.66666f;
						}
					}
					
						
					player.GetComponent<Player>().GenDynamicMesh(pivot_from, out pivot_offset_vector_x, out pivot_offset_vector_y);
					player.transform.position = pivot_from;
					
					//calculate pivot axis
					
					Matrix4x4 M = transform.localToWorldMatrix;
					Matrix4x4 V = cam.worldToCameraMatrix;
//					Matrix4x4 P = cam.projectionMatrix;				
					Matrix4x4 MV = V * M;										
					//Matrix4x4 V = V;		
					Vector3 movingfrom_dir_projected = MV.MultiplyVector(movingfrom_dir);
					
					Vector3 flattened1 = new Vector3(movingfrom_dir_projected.x, movingfrom_dir_projected.y, 0);
					Vector3 flattened2 = new Vector3(pivot_offset_vector_y.x, pivot_offset_vector_y.y, 0);
					pivot_axis = Vector3.Cross(flattened2, flattened1).normalized;
							
					player.transform.position -= camforward * 50;
					
					SetPieceMask(level.playerface, 0, f);
					SetPieceMask(f, 1, level.playerface);
					
					level.playerface = f;
					
					Face temp;
					level.ProjectFaceOut(f, mouseorbit.targetview, out temp);
					
					level.TryCollect(f);
					//SnapPlayer();
				}
			
			}
		}
			
		
		float oldlerptime = lerptime;
		lerptime += Time.deltaTime / movespeed;
		if (oldlerptime < lerplength / 2 && lerptime >= lerplength / 2)
		{
			//Debug.DrawLine(pivot_from, pivot_to, Color.yellow, 2);
			player.transform.position += pivot_to - pivot_from;
		}
		
		if (oldlerptime <= lerplength && lerptime > lerplength)
		{
			SnapPlayer();
			HidePieceMasks();			
			level.UpdateCollectableCount();
			animdone = true;
		}
		if (lerptime < lerplength)
		{		
			
			float angle = 180.0f * Time.deltaTime / movespeed;
			Quaternion q = Quaternion.AngleAxis(angle, pivot_axis);
			pivot_offset_vector_x = q * pivot_offset_vector_x;
			pivot_offset_vector_y = q * pivot_offset_vector_y;
			player.renderer.material.SetVector("_ByX", pivot_offset_vector_x);			
			player.renderer.material.SetVector("_ByY", pivot_offset_vector_y);			
			
			/*
			float angle = 180.0f * Time.deltaTime / movespeed;
			if (lerptime < lerplength / 2)
			{
				player.transform.RotateAround(pivot_from, pivot_axis, angle);
			}
			else
			{
				player.transform.RotateAround(pivot_to, pivot_axis, angle);
			}*/
		}
		/*
		if (lerptime < 0.5)
		{			
			float t = lerptime * 2;
			player.transform.position = source_from * (1 - t) + source_via * (t);
			player.renderer.material.color = new Color(1, 1, 1, 1 - lerptime);
		}
		else
		if (lerptime < 1.0f)
		{
			float t = (lerptime - 0.5f) * 2;
			player.transform.position = target_via * (1 - t) + target_to * (t);
			player.renderer.material.color = new Color(1, 1, 1, lerptime);			
		}
		else
		{
			SnapPlayer();
		}	*/	
	}
	
	public float dilation=1.1f;
	
	public float hoverheight = 1.05f;
	public void SnapPlayer ()
	{
		int s = level.playerface.side;
		player.transform.position = level.playerface.PushedCenter(hoverheight);
		player.transform.rotation = Quaternion.LookRotation(DodecaData.SideForward[s], DodecaData.SideNormal[s]);
		player.GetComponent<Player>().GenStaticMesh();
	}
	
	private void SetPieceMask (Face f, int index, Face otherface)
	{
		Vector3 camforward = -DodecaData.vertices [mouseorbit.targetview];
		
		Face obscuring;
		if (FullyVisible(f, out obscuring, mouseorbit.targetview) == false)
		{
			SetMaskDodeca(obscuring.coord, index);
						
			//raycast to find any adjoining faces with the same side (necessarily of a different parent)
		
			int k = 0;
			Vector3[] edgedirs = DodecaData.EdgeDirs [obscuring.side];		
			foreach (Vector3 v in edgedirs)
			{							
				Vector3 fullway = obscuring.Center() + v * 1.1f;
				Ray fullwayray = new Ray(fullway - 100 * camforward, camforward);
				RaycastHit hit;
				if (Physics.Raycast(fullwayray, out hit, 1000, 1 << 10))
				{
					Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
					if (newface.side == obscuring.side)
					{
						SetMaskTile(index * masklength + 3 + k, newface);
						k++;
						
						if (newface.side != otherface.side)
						{
							//messy second level of recursion.  bite me.
							
							foreach (Vector3 v2 in edgedirs)
							{		
								if (v2 == v)
								{
									continue;
								}
								
								Vector3 fullway2 = newface.Center() + v2 * 1.1f;
								Ray fullwayray2 = new Ray(fullway2 - 100 * camforward, camforward);
								RaycastHit hit2;
								if (Physics.Raycast(fullwayray2, out hit2, 1000, 1 << 10))
								{
									Face newface2 = level.GetFaceFromNormalRay(hit2.point, hit2.normal);
									if (newface2.side == obscuring.side)
									{
										SetMaskTile(index * masklength + 3 + k, newface2);
										k++;										
									}
								}			
							}
						}
					}
				}			
			}
		}
		else
		{
			for (int i=0; i<masklength; i++)
			{
				piecemask [masklength * index + i].renderer.enabled = false;
			}
		}
		
	}
	
	private void HidePieceMasks ()
	{
		for (int i = 0; i<maskcount; i++)
		{
			piecemask [i].renderer.enabled = false;
		}
	}
	
		
	private void SetMaskDodeca (Coord c, int index)
	{
		foreach (Dodeca d in level.dodecas)
		{
			if (d.coord.ValueEquals(c))
			{
				//a good un :)
				int i = 0;
				foreach (Face f in d.faces)
				{
					if (AtLeastPartiallyVisible(f,mouseorbit.targetview))
					{
						SetMaskTile(index * masklength + i, f);
						i++;
					}
				}
			}
		}
	}
	

	private void SetMaskTile (int i, Face f)
	{
		Vector3 camforward = - DodecaData.vertices[mouseorbit.targetview];
		piecemask [i].renderer.enabled = true;
		int s = f.side;
		piecemask [i].transform.position = f.Center() - 100 * camforward;
		piecemask [i].transform.rotation = Quaternion.LookRotation(DodecaData.SideForward [s], DodecaData.SideNormal [s]);
	}
}
