using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

public struct Coord {
		
	public static Coord Zero = new Coord(0,0,0);
	
	public readonly int x;
	public readonly int y;
	public readonly int z;
	
	public string Serialize ()
	{
		return x + "," + y + "," + z;
	}
	
	public static Coord Deserialize (string s)
	{
		string[] split = s.Split(',');
		int x = int.Parse(split [0]);
		int y = int.Parse(split [1]);
		int z = int.Parse(split [2]);
		return new Coord(x,y,z);
	}
	
	public bool ValueEquals (Coord other)
	{
		return x==other.x && y== other.y && z== other.z;
	}
	

	public Coord (int _x, int _y, int _z)
	{
		x = _x;
		y = _y;
		z = _z;		
	}
	
	public string Print ()
	{
		return x + "," + y + "," + z;
	}
	
	//NODE assumes point is of degree 3
	public Coord GenerateNeighbourToPointOfDegree3 (int p)
	{
		int[] sides = DodecaData.NeighbouringSidesOfVertex [p];
		return this
			.GetNeighbourToSide(sides [0])
			.GetNeighbourToSide(sides [1])
			.GetNeighbourToSide(sides [2]);
	}
	
	
	//gets neighbour to side
	public Coord GetNeighbourToSide (int s)
	{
		int dx = x % 2 == 0 ? 1 : 0;		
		int dy = y % 2 == 0 ? 1 : 0;
		
		switch (s)
		{
			case 0:
				return new Coord(2 * dy + x - 1, y + 1, z - 1 + dx);
			case 1:
				return new Coord(2 * dy + x, y + 1, z);
			case 2:
				return new Coord(2 * dy + x - 1, y + 1, z + dx);
			case 3:
				return new Coord(2 * dy + x - 2, y + 1, z);
			case 4:
				return new Coord(x - 1, y, dx + z - 1);
			case 5:
				return new Coord(x + 1, y, dx + z - 1);
			case 6:
				return new Coord(x + 1, y, dx + z);
			case 7:
				return new Coord(x - 1, y, dx + z);
			case 8:
				return new Coord(2 * dy + x - 1, y - 1, z - 1 + dx);
			case 9:
				return new Coord(2 * dy + x, y - 1, z);
			case 10:
				return new Coord(2 * dy + x - 1, y - 1, z + dx);
			case 11:
				return new Coord(2 * dy + x - 2, y - 1, z);
			default:
				Debug.LogError("ERROR no vallid side " + s);
				return default(Coord);
		}
	}
	
	//disjoint from regular neighbourhood
	public Coord[] VertexTouchingNeighbourhood ()
	{
		Coord[] neighbours = new Coord[6];
		
		for (int j=0; j<6; j++)
		{
			int d = DodecaData.vertexneighbourdirs [j];
			int[] sides = DodecaData.NeighbouringSidesOfVertex [d];
			//assumes a full four
			//to get neighbouring side, need to translate by two opposite sides.
			Vector3 s1 = DodecaData.SideNormal [sides [0]];
			int s1_index = sides[0];
			int s2_index = 0;
			for (int i=1; i<4; i++)
			{
				Vector3 cand_s = DodecaData.SideNormal [sides [i]];
				float a = Vector3.Angle(cand_s, s1);
				if (a > 80)
				{
					s2_index = sides [i];
					break;
				}
			}
			
			neighbours [j] = Translate(s1_index).Translate(s2_index);
			
		}
		
		return neighbours;
	}
	
	public Coord[] Neighbourhood ()
	{	
		Coord[] neighbours = new Coord[12];

		int dx = x % 2 == 0 ? 1 : 0;
		int dy = y % 2 == 0 ? 1 : 0;
		
		
		//UPPER
		neighbours [0] = new Coord(2 * dy + x - 1, y + 1, z - 1 + dx);
		neighbours [1] = new Coord(2 * dy + x, y + 1, z);
		neighbours [2] = new Coord(2 * dy + x - 1, y + 1, z + dx);
		neighbours [3] = new Coord(2 * dy + x - 2, y + 1, z);
		neighbours [4] = new Coord(x - 1, y, dx + z - 1);
		neighbours [5] = new Coord(x + 1, y, dx + z - 1);
		neighbours [6] = new Coord(x + 1, y, dx + z);
		neighbours [7] = new Coord(x - 1, y, dx + z);
		neighbours [8] = new Coord(2 * dy + x - 1, y - 1, z - 1 + dx);
		neighbours [9] = new Coord(2 * dy + x, y - 1, z);
		neighbours [10] = new Coord(2 * dy + x - 1, y - 1, z + dx);
		neighbours [11] = new Coord(2 * dy + x - 2, y - 1, z);
			
		return neighbours;
		
	}
	
	public Vector3 ToVector3 ()
	{
		Vector3 pos = new Vector3(2*x, 2*y, 4*z);
					
		if (x % 2 == 0)
		{
			pos = pos + new Vector3(0, 0, 2.0f);
		}
		if (y % 2 == 0)
		{
			pos = pos + new Vector3(2.0f, 0, 0);
		}
		
		return pos;
	}
	
	
	public static Coord FromVector3 (Vector3 v)
	{
		int x = Mathf.RoundToInt(v.x/2.0f);
		int y = Mathf.RoundToInt(v.y / 2.0f);
		int z = Mathf.RoundToInt(v.z / 2.0f);
		
		if (y % 2 == 0)
		{
			x -= 1;
		}
		
		if (x % 2 == 0)
		{
			z -=1;
		}
		
		return new Coord(x, y, Mathf.RoundToInt(z/2.0f));				
	}
	/*
	public static Coord FromVector3 (Vector3 pos)
	{		
		int x = Mathf.RoundToInt(pos.x);
		int y = Mathf.RoundToInt(pos.y);
		int z = Mathf.RoundToInt(pos.z);
		
		if (y % 2 == 0)
		{
			x -= 2;
		}
		if (x % 2 == 0)
		{
			z -= 2;
		}
		return new Coord(x/2, y/2, z/2);
	}*/

	public static bool Neighbours (Coord a, Coord b)
	{
		var neighbours = a.Neighbourhood();
		foreach (Coord c in neighbours)
		{
			if (c.ValueEquals(b))
			{
				return true;
			}			
		}
		return false;
	}
	
	public Coord Translate (int dir)
	{
		return Neighbourhood()[dir];
	}
}
