using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public struct Face {
		
	public readonly Coord coord;
	public readonly int side;
	
	public string Serialize ()
	{
		return coord.Serialize()+"^"+(int)side;
	}
	
	public static Face Deserialize (string s)
	{
		string[] split = s.Split('^');
		Coord coord = Coord.Deserialize(split [0]);
		int i = int.Parse(split [1]);
		int side = i;
		return new Face(coord, side);
	}
	
	public bool ValueEquals (Face other)
	{
		return side==other.side && coord.ValueEquals(other.coord);
	}
	
	public string Print ()
	{
		return "("+coord.Print()+" - " + side.ToString()+")";
	}
	
	public Face (Coord _coord, int _side)
	{
		coord = _coord;
		side = _side;
	}	
	
	public Vector3 Center ()
	{
		return coord.ToVector3() + DodecaData.SideNormal[side];
	}
	
	public Vector3 PushedCenter ()
	{
		return PushedCenter(1.025f);
	}
	
	public Vector3 PushedCenter(float f)
	{
		return coord.ToVector3() + DodecaData.SideNormal[side]*f;
	}
	
	public Coord NormalCoord ()
	{
		return coord.Translate(side);
	}
	
	public List<Vector3> EdgeDirs ()
	{
		List<Vector3> result = new List<Vector3>();
		int[] dirs = DodecaData.NeighbouringDirs[(int)side];
		
		foreach (int o in dirs)
		{
			Vector3 on = DodecaData.SideNormal[o];
			Vector3 v = (on - Vector3.Project(on, DodecaData.SideNormal[side])).normalized;
			result.Add(v);
		}
		return result;			
	}	
	
	
	private static int[] indices = {3,1,2,0};
	
	public List<Vector3> RealEdgeDirs ()
	{
		List<Vector3> result = new List<Vector3>();
		Vector3[] quadpoints = _QuadPointsOnly();
		
		Vector3 center = Center();
				
		for (int i=0; i<4; i++)
		{
			Vector3 v1 = quadpoints [indices [i]];
			Vector3 v2 = quadpoints [indices [(i + 1) % 4]];
			Vector3 v_avg = (v1 + v2) / 2;
			result.Add(v_avg - center);
		}
		
		return result;
	}	
	
	private Vector3[] _QuadPointsOnly ()
	{
		Face f = this;
		Vector3 pos = new Vector3(2 * f.coord.x, 2 * f.coord.y, 4 * f.coord.z);
		if (f.coord.x % 2 == 0)
		{
			pos = pos + new Vector3(0, 0, 2);
		}
		if (f.coord.y % 2 == 0)
		{
			pos = pos + new Vector3(2, 0, 0);
			//pos.x+=1;
		}
		Vector3[] points = new Vector3[4];
		for (int i = 0; i < 4; i++)
		{
			int j = DodecaData.dodec_tris [((int)f.side) * 4 + i];
			points [i] = DodecaData.vertices [j] + pos;
		}
		return points;
	}
	
	public Vector3[] QuadPoints ()
	{
		Face f = this;
		Vector3 pos = new Vector3(2 * f.coord.x, 2 * f.coord.y, 4 * f.coord.z);
		if (f.coord.x % 2 == 0)
		{
			pos = pos + new Vector3(0, 0, 2);
		}
		if (f.coord.y % 2 == 0)
		{
			pos = pos + new Vector3(2, 0, 0);
			//pos.x+=1;
		}
		Vector3[] points = new Vector3[8];
		for (int i = 0; i < 4; i++)
		{
			int k = ((int)f.side) * 4 + i;
			int j = DodecaData.dodec_tris [k];
			points [i] = DodecaData.vertices [j] + pos;
		}
		
		{
			Vector3 uv = DodecaData.SideUV [(int)f.side];
			points [4] = uv;
			points [5] = uv;
			points [6] = uv;
			points [7] = uv;
		}
		
		return points;
	}
	
}
