using UnityEngine;
using System.Collections.Generic;
using System.Collections;
using System;

public class Level : MonoBehaviour {
	public List<Coord> points;
	public List<Dodeca> dodecas;
	public Face playerface;
	
	public bool leaving=false;
	public List<Face> collectables;
	public List<GameObject> collectable_obs;
	public Camera cam;
	
	public GameObject text;
	
	public int initcollectablecount;
	
	public string Serialize ()
	{
		//pointcount|points...,collecatablecount,..collectables,playerpos,cameradir,targetup
		string result = "";
		result += points.Count.ToString() + "|";
		foreach (Coord c in points)
		{
			result += c.Serialize() + "|";
		}
		
		result += collectables.Count.ToString() + "|";
		
		foreach (Face f in collectables)
		{
			result += f.Serialize() + "|";
		}
		
		result += playerface.Serialize() + "|";
		
		result += mo.targetview.ToString() + "|";
		result += Utility.SerializeVector(mo.transform.up);
		
		return result;
	}
	//not actually the centroid any more - OH WELL
	public Vector3 Centroid ()
	{
		float maxx = -1000;
		float minx = 1000;
		float maxy = -1000;
		float miny = 1000;
		float maxz = -1000;
		float minz = 1000;
		
		foreach (Coord p in points)
		{
			Vector3 v = p.ToVector3();
			maxx = Mathf.Max(maxx, v.x);
			maxy = Mathf.Max(maxy, v.y);
			maxz = Mathf.Max(maxz, v.z);
			minx = Mathf.Min(minx, v.x);
			miny = Mathf.Min(miny, v.y);
			minz = Mathf.Min(minz, v.z);
		}
		return new Vector3((maxx + minx) / 2, (maxy + miny) / 2, (maxz + minz) / 2);
	}
		
	public void Jitter ()
	{
		bool modified = false;
		float min = Centroid().magnitude;
		int c_x = 0;
		int c_y = 0;
		int c_z = 0;
		int r = 2;
		for (int x=-r; x<=r; x++)
		{
			for (int y=-r; y<=r; y++)
			{
				for (int z=-r; z<=r; z++)
				{
					Translate(2*x, 2*y, z);
					float dist = Centroid().magnitude;
					if (dist < min)
					{
						min = dist;
						modified = true;
						c_x = 2*x;
						c_y = 2*y;
						c_z = z;
					}
					Translate(-2*x, -2*y, -z);
				}
			}
		}			
				
		if (modified)
		{
			Debug.Log("jittering by " + c_x + "," + c_y + "," + c_z);
			Translate(c_x, c_y, c_z);
			Jitter();
		}

	}
	public void Translate (int x, int y, int z)
	{		
		for (int i=0; i<points.Count; i++)
		{
			Coord c = points [i];
			points [i] = new  Coord(c.x + x, c.y + y, c.z + z);
		}
		
		for (int i=0; i<collectables.Count; i++)
		{
			Face f = collectables [i];
			Coord c = f.coord;
			collectables [i] = new Face(new  Coord(c.x + x, c.y + y, c.z + z), f.side);
		}
		
		{
			Coord c = playerface.coord;
			playerface = new Face(new  Coord(c.x + x, c.y + y, c.z + z), playerface.side);
		}
	}
	

	public void Deserialize (string dat)
	{
		ClearCollectables();
		
		//pointcount|points...,collecatablecount,..collectables,playerpos,cameradir,targetup
		string[] dats = dat.Split('|');

		int i = 0;
		int pointcount = int.Parse(dats [0]);
		points = new List<Coord>();
				
		for (int j=0; j<pointcount; j++)
		{
			i++;
			Coord c = Coord.Deserialize(dats [i]);
			points.Add(c);
		}
		
		i++;
		initcollectablecount = int.Parse(dats [i]);
		
		collectables = new List<Face>();
		for (int j=0; j<initcollectablecount; j++)
		{
			i++;
			Face f = Face.Deserialize(dats [i]);
			collectables.Add(f);
		}
		
		i++;
		playerface = Face.Deserialize(dats [i]);
		
		i++;
		mo.targetview = int.Parse(dats [i]);		
		
		i++;
		mo.targetup = Utility.DeserializeVector(dats [i]);			
		
		InstantiateLevel();
	}
	
	
	public static void CullEmptyDodecas (List<Dodeca> dodecas, List<Coord> points)
	{
		for (int i=dodecas.Count-1; i>=0; i--)
		{
			Dodeca dodeca = dodecas [i];
			if (Blocked(dodeca.coord, points))
			{
				dodecas.RemoveAt(i);
			}
		}		
	}
	
	public static bool Blocked (Coord c, List<Coord> points)
	{
		return NeighboursOf(c, points).Count == 12;
	}
	
	public static List<Coord> NeighboursOf (Coord c, List<Coord> points)
	{
		List<Coord> result = new List<Coord>();
		
		Coord[] candneighbours = c.Neighbourhood();
		
		foreach (Coord coord in candneighbours)
		{
			if (points.IndexOf(coord) >= 0)
			{
				result.Add(coord);
			}
		}
		return result;
	}
	
	
	public static List<Coord> VertexConnectedNeighboursOf (Coord c, List<Coord> points)
	{
		List<Coord> result = NeighboursOf(c, points);
		
		Coord[] candneighbours = c.VertexTouchingNeighbourhood();
		
		foreach (Coord coord in candneighbours)
		{
			if (points.IndexOf(coord) >= 0)
			{
				result.Add(coord);
			}
		}
		return result;
	}
	
	public static void CullBlockedFaces (List<Dodeca> dodecas, List<Coord> points)
	{
		for (int i=0; i<dodecas.Count; i++)
		{
			Dodeca d = dodecas [i];
			Coord c = d.coord;
			List<Coord> neighbours = NeighboursOf(d.coord, points);
			
			for (int j=d.faces.Count-1; j>=0; j--)
			{
				Face f = d.faces [j];
				if (neighbours.IndexOf(c.Translate(f.side)) >= 0)
				{
					d.faces.RemoveAt(j);
				}
			}		
		}		
	}
	
	
	private void InstantiateLevel ()
	{
		//generate mesh		
		dodecas = new List<Dodeca>();
		foreach (Coord p in points)
		{
			dodecas.Add(new Dodeca(p));
		}		
		
		CullEmptyDodecas(dodecas, points);
		CullBlockedFaces(dodecas, points);
		GenerateMesh();
		
		//set player position
		controller.Reset();
		controller.SnapPlayer();
		
		//spawn collectables
		SpawnCollectables();
		
		//set camera position+state
		mo.SnapWithUp();
		
		UpdateCollectableCount();
		
		//PrintLevelStats();
	}
	
	//viewdir indexes into vertices - assuming a valid one has been picked
	public bool ProjectFaceOut (Face f, int viewdir, out Face result)
	{
		Vector3 viewvector = - DodecaData.vertices [viewdir];		
		
		Vector3[] edgedirs = DodecaData.EdgeDirs [(int)f.side];
		foreach (Vector3 v in edgedirs)
		{
			Vector3 halfway = f.Center() + v * 1.1f / 2.0f;
			Ray halfwayray = new Ray(halfway - 100 * viewvector, viewvector);
			RaycastHit hit;
			if (Physics.Raycast(halfwayray, out hit, 1000, 1 << 10))
			{
				Face newface = GetFaceFromNormalRay(hit.point, hit.normal);
				if (newface.side == f.side)
				{
					result = newface;
					
					return true;
				}
			}
		}
		
		result = default(Face);
		return false;
	}
	/*
	//viewdir indexes into vertices - assuming a valid one has been picked
	public bool ProjectFaceOutNew (Face f, int viewdir, out Face result)
	{
		Coord fbase = f.coord;		
		Coord nearest = fbase;
		bool blockedL = false;
		bool blockedR = false;
		int[] sides = DodecaData.NeighbouringSidesOfVertex [viewdir];
				
		int sx;
		int sy;
		if (f.side == sides [0])
		{
			sx = sides [1];
			sy = sides [2];
		}
		else
		if (f.side == sides [1])
		{
			sx = sides [0];
			sy = sides [2];
		}
		else
		{
			sx = sides [0];
			sy = sides [1];			
		}
		
		
		//i *think* randomdiameter is good ehre.
		for (int i=0; i<randomdiameter; i++)
		{
			Coord obstructor = fbase.GetNeighbourToSide(f.side);
			//check if blocked by obstructor
			if (points.IndexOf(obstructor) >= 0)
			{
				blockedL = true;
				blockedR = true;
			}
			
			Coord ob1 = obstructor.Translate(sx);
			Coord ob2 = obstructor.Translate(sy);
			if (points.IndexOf(ob1) >= 0)
			{
				blockedL = true;					
			}
			
			if (points.IndexOf(ob2) >= 0)
			{
				blockedR = true;									
			}
			
			//check if blocked by two neighbours of obstructor
			fbase = fbase.GenerateNeighbourToPointOfDegree3(viewdir);
			
			if (points.IndexOf(fbase) >= 0)
			{
				nearest = fbase;
				blockedL = false;
				blockedR = false;
			}
		}
		
		if (blockedL && blockedR)
		{
			result = default(Face);
			return false;
		}
		else
		{
			result = new Face(nearest, f.side);
			return true;
		}
	}
	*/
	/*
	public int trinketcount = 6;
	
	
	public void PlacePlayerAndCollecatbles ()
	{
		//List<Face> accessiblefaces = AccessibleFaces();
		dodecas.Shuffle();
		
		ClearCollectables();
		collectables = new List<Face>();

		
		for (int i=0; i<trinketcount; i++)
		{
			dodecas [i % dodecas.Count].faces.Shuffle();
			collectables.Add(dodecas [i % dodecas.Count].faces [0]);
		}
		dodecas [trinketcount % dodecas.Count].faces.Shuffle();
		playerface = dodecas [trinketcount % dodecas.Count].faces [0];
		
		GetComponent<Controller>().SnapPlayer();
		mo.Target();
		SpawnCollectables();
		mo.Snap();
		ShowInvisibleCollectables();
		HideInvisibleCollectables();	
	}*/
	/*
	private void PrintLevelStats ()
	{
		List<Face> faces = Faces();
		List<Face> accessiblefaces = AccessibleFaces();
		List<List<Face>> connectedcomponents = ConnectedComponents(accessiblefaces);
		List<List<Coord>> connectedpointcomponents = ConnectedPointComponents();
		List<List<Coord>> vertexconnectedpointcomponents = VertexConnectedPointComponents();
		float deadendchance;
		float averagenumberofdirs;
		int deg8faces;
		float univiewablefacechance;
		int parallelmovements;
		float chancehalfcovered;
		float chancehalfbicovered;
		FaceStatsAggregate(accessiblefaces, out deadendchance, out averagenumberofdirs, out deg8faces, out univiewablefacechance, out parallelmovements, out chancehalfcovered, out chancehalfbicovered);
		
		int facecount = faces.Count;
		int accessiblefacecount = accessiblefaces.Count;
		int inaccessiblefacecount = facecount - accessiblefacecount;
		//int connectedcomponentcount = connectedcomponents.Count;
		//int connectedpointcomponentcount = connectedpointcomponents.Count;
		//int vertexconnectedpointcomponentcount = vertexconnectedpointcomponents.Count;
		
		string r = "";
		
		r += "Faces " + Faces().Count + " (" + accessiblefacecount + "|" + inaccessiblefacecount + ")\n";
				
		r += "Components " + connectedcomponents.Count + " (";
		foreach (List<Face> f in connectedcomponents)
		{
			r += f.Count + ",";
		}
		r = r.Substring(0, r.Length - 1) + ")\n";
				
		r += "Point Components " + connectedpointcomponents.Count + " (";
		foreach (List<Coord> c in connectedpointcomponents)
		{
			r += c.Count + ",";
		}
		r = r.Substring(0, r.Length - 1) + ")\n";
				
		r += "Vertex-connected Point Components " + vertexconnectedpointcomponents.Count + " (";
		foreach (List<Coord> c in vertexconnectedpointcomponents)
		{
			r += c.Count + ",";
		}
		r = r.Substring(0, r.Length - 1) + ")\n";
			
		
		r += "dead end chance = " + deadendchance + "\n";
		r += "# average moveable dirs = " + averagenumberofdirs + "\n";
		r += "# deg 8 faces = " + deg8faces + "\n";
		r += "% univiewable faces " + univiewablefacechance + "\n";
		r += "# parallel moves = " + parallelmovements + "\n";
		r += "% half-covered = " + chancehalfcovered + "\n";
		r += "% half-bi-covered = " + chancehalfbicovered + "\n";
		Debug.Log(r);
		
	}*/
	/*
	void OnDrawGizmos ()
	{
		if (vertexconnectedpointcomponents != null)
		{

			for (int i=0; i<vertexconnectedpointcomponents.Count; i++)
			{
				Color c = (new HSBColor((float)i / vertexconnectedpointcomponents.Count, 1.0f, 0.6f)).ToColor();
				Gizmos.color = c;
				for (int j=0; j<vertexconnectedpointcomponents[i].Count; j++)
				{
					Coord coord = vertexconnectedpointcomponents [i] [j];
					Gizmos.DrawWireSphere(coord.ToVector3(), 1);
					//Debug.DrawRay(f.Center(), DodecaData.SideNormal [f.side], c, 5);
				}
			}
		}
	}*/

	private List<Face> Faces ()
	{
		List<Face> faces = new List<Face>();
		foreach (Dodeca d in dodecas)
		{
			foreach (Face f in d.faces)
			{
				faces.Add(f);
			}
		}
		return faces;
	}
	/*
	private List<Face> AccessibleFaces ()
	{
		List<Face> allfaces = Faces();
		List<Face> result = new List<Face>();

		foreach (Face f in allfaces)
		{
			bool visible = false;
			foreach (int viewdir in DodecaData.sideviewdirs[f.side])
			{
				Face f_p1;
				bool result1 = ProjectFaceOut(f, viewdir, out f_p1);

				
				if (result1)
				{
					if (f_p1.ValueEquals(f))
					{
						visible = true;
						break;
					}
				}
			}
					
			if (visible)
			{
				result.Add(f);
			}
		}
		
		return result;		
	}
	
	private int GetGroup (Face f, List<List<Face>> faces)
	{
		for (int i=0; i<faces.Count; i++)
		{
			for (int j=0; j<faces[i].Count; j++)
			{
				if (f.ValueEquals(faces [i] [j]))
				{
					return i;
				}
			}
		}
		return -1;
	}
	
	
	private int GetCoordGroup (Coord c, List<List<Coord>> coords)
	{
		for (int i=0; i<coords.Count; i++)
		{
			for (int j=0; j<coords[i].Count; j++)
			{
				if (c.ValueEquals(coords [i] [j]))
				{
					return i;
				}
			}
		}
		return -1;
	}*/
	/*
	private void FaceStatsAggregate (List<Face> accessiblefaces, out float deadendchance, out float movedircountaverage, out int deg8faces, out float univiewablefacechance, out int parallelmovements, out float chancehalfcovered, out float chancebihalfcovered)
	{
		deadendchance = 0;
		movedircountaverage = 0;
		deg8faces = 0;
		univiewablefacechance = 0;
		parallelmovements = 0;
		chancehalfcovered = 0;
		chancebihalfcovered = 0;
		foreach (Face f in accessiblefaces)
		{			
			int totaloutbound = 0;
			bool univiewable = false;
			int partiallyvisible = 0;
			foreach (int possibleviewdir in DodecaData.sideviewdirs[f.side])
			{
				if (controller.ExactlyPartiallyVisible(f, possibleviewdir))
				{
					partiallyvisible++;
				}
				
				List<Face> adjacent = controller.GetAllowedFaces(f, possibleviewdir);
				foreach (Face o in adjacent)
				{
					if (o.side == f.side)
					{
						parallelmovements++;
					}
				}
				movedircountaverage += adjacent.Count;
				totaloutbound += adjacent.Count;

				if (adjacent.Count == 0)
				{
					univiewable = true;
				}
			}
			if (partiallyvisible > 0)
			{
				chancehalfcovered++;
			}
			if (partiallyvisible == 2)
			{
				chancebihalfcovered++;
			}
			
			if (univiewable)
			{
				univiewablefacechance++;
			}			
			if (totaloutbound == 8)
			{
				deg8faces++;			
			}
			else
			if (totaloutbound == 2)
			{
				deadendchance++;
			}
		}
		
		chancehalfcovered /= accessiblefaces.Count;
		chancebihalfcovered /= accessiblefaces.Count;
		
		deadendchance /= accessiblefaces.Count;
		parallelmovements /= 2;//double-counting
		movedircountaverage /= (2 * accessiblefaces.Count);//from a particular view
		univiewablefacechance /= (accessiblefaces.Count);
	}
	
	private List<List<Face>> ConnectedComponents (List<Face> accessiblefaces)
	{		
		List<List<Face>> result = new List<List<Face>>();
		
		foreach (Face f in accessiblefaces)
		{	
			List<int> groups = new List<int>();
			
			//go through adjacent squares, see if any already referenced
			foreach (int possibleviewdir in DodecaData.sideviewdirs[f.side])
			{
				List<Face> adjacent = controller.GetAllowedFaces(f, possibleviewdir);
				foreach (Face other in adjacent)
				{
					int facegroup = GetGroup(other, result);
					if (facegroup >= 0 && groups.IndexOf(facegroup) == -1)
					{
						groups.Add(facegroup);
					}
				}
			}
			
			//if none referenced, add to new group
			if (groups.Count == 0)
			{
				List<Face> newgroup = new List<Face>();
				newgroup.Add(f);
				result.Add(newgroup);
			}
			//if just one referenced, add to that group
			else
			if (groups.Count == 1)
			{
				int g = groups [0];
				result [g].Add(f);
			}
			//if more than one referenced, merge the groups and add yourself to the result
			else
			{
				List<Face> newgroup = new List<Face>();	
				newgroup.Add(f);
				groups.Sort();
				for (int i=groups.Count-1; i>=0; i--)
				{
					int g = groups [i];
					newgroup.AddRange(result [g]);
					result.RemoveAt(g);
				}				
				result.Add(newgroup);
			}
		}
		return result;		
	}
	
	
	private List<List<Coord>> ConnectedPointComponents ()
	{
		List<List<Coord>> result = new List<List<Coord>>();
		
		foreach (Coord p in points)
		{	
			List<int> groups = new List<int>();
			
			List<Coord> neighbours = NeighboursOf(p, points);
			//go through adjacent squares, see if any already referenced
			foreach (Coord other in neighbours)
			{
				int facegroup = GetCoordGroup(other, result);
				if (facegroup >= 0 && groups.IndexOf(facegroup) == -1)
				{
					groups.Add(facegroup);
				}
			}
			
			
			//if none referenced, add to new group
			if (groups.Count == 0)
			{
				List<Coord> newgroup = new List<Coord>();
				newgroup.Add(p);
				result.Add(newgroup);
			}
			//if just one referenced, add to that group
			else
			if (groups.Count == 1)
			{
				int g = groups [0];
				result [g].Add(p);
			}
			//if more than one referenced, merge the groups and add yourself to the result
			else
			{
				List<Coord> newgroup = new List<Coord>();	
				newgroup.Add(p);
				groups.Sort();
				for (int i=groups.Count-1; i>=0; i--)
				{
					int g = groups [i];
					newgroup.AddRange(result [g]);
					result.RemoveAt(g);
				}		
				result.Add(newgroup);
			}
		}
		return result;		
	}
	
	
	
	
	private List<List<Coord>> VertexConnectedPointComponents ()
	{
		List<List<Coord>> result = new List<List<Coord>>();
		
		foreach (Coord p in points)
		{	
			List<int> groups = new List<int>();
			
			List<Coord> neighbours = VertexConnectedNeighboursOf(p, points);
			//go through adjacent squares, see if any already referenced
			foreach (Coord other in neighbours)
			{
				int facegroup = GetCoordGroup(other, result);
				if (facegroup >= 0 && groups.IndexOf(facegroup) == -1)
				{
					groups.Add(facegroup);
				}
			}
			
			
			//if none referenced, add to new group
			if (groups.Count == 0)
			{
				List<Coord> newgroup = new List<Coord>();
				newgroup.Add(p);
				result.Add(newgroup);
			}
			//if just one referenced, add to that group
			else
			if (groups.Count == 1)
			{
				int g = groups [0];
				result [g].Add(p);
			}
			//if more than one referenced, merge the groups and add yourself to the result
			else
			{
				List<Coord> newgroup = new List<Coord>();	
				newgroup.Add(p);
				groups.Sort();
				for (int i=groups.Count-1; i>=0; i--)
				{
					int g = groups [i];
					newgroup.AddRange(result [g]);
					result.RemoveAt(g);
				}		
				result.Add(newgroup);
			}
		}
		return result;		
	}*/
	
	MouseOrbit mo;
	Controller controller;
	public void Awake ()
	{
		controller = GetComponent<Controller>();
		
		cam = Camera.mainCamera;
		mo = cam.GetComponent<MouseOrbit>();
		
		GetComponent<MeshFilter>().sharedMesh = new Mesh();
		points = new List<Coord>();
		dodecas = new List<Dodeca>();
		collectables = new List<Face>();
		collectable_obs = new List<GameObject>();
		/*
		PopulateWithRandomPoints();
		Regen();
		GenerateCollectables(4);
		*/
		//RandomLevel();
		Deserialize(LevelData.LevelDats [TitleOptions.curlevel]);
		/*
		
		float h = (float)LevelSelect.targetlevel / 14.0f;
		HSBColor hsb = new HSBColor(h, 0.4f, 0.8f);
		cam.backgroundColor = hsb.ToColor();
		RenderSettings.fogColor = cam.backgroundColor;*/
	}
	
	/*
	public void RandomLevel ()
	{
		ClearCollectables();
		
		//GetComponent<MeshFilter>().sharedMesh = new Mesh();
		points = new List<Coord>();
		dodecas = new List<Dodeca>();
		collectables = new List<Face>();
		collectable_obs = new List<GameObject>();
		
		PopulateWithRandomPoints();
		Regen();
		GenerateCollectables(4);
		Jitter();
		Reload();
	}
	
	
	public void FastRandomLevel ()
	{
		ClearCollectables();
		
		//GetComponent<MeshFilter>().sharedMesh = new Mesh();
		points = new List<Coord>();
		dodecas = new List<Dodeca>();
		collectables = new List<Face>();
		collectable_obs = new List<GameObject>();
		
		PopulateWithRandomPoints();
		Regen();
	}
	*/
	
	public void Reload ()
	{
		Deserialize(Serialize());
	}
	public int failthreshold = 1200;
	
	/*
	private IEnumerator GenerateInteresting ()
	{
		Application.runInBackground = true;
		//System.Diagnostics.Stopwatch sw = System.Diagnostics.Stopwatch.StartNew();
		Debug.Log("generating interesting");
		
		float[] track = {-1.0f,-1.0f,10000,-1.0f,-1.0f,-1.0f};
		string[] trackdescrs = {"deadends","vertex-disconnected chunks","min avr exits","high univiewable faces","half-covered","bi-half-covered"};
		string[] tracklevels = {"","","","","",""};
		
		
		//for (int i=0; i<200; i++)
		int failcount = 0;
		for (int i=0; true; i++)
		{
			//randompoints = UnityEngine.Random.Range(7, 30);
			if (failcount > failthreshold || Input.GetKeyDown(KeyCode.Y))
			{
				failcount = 0;
				
				string d = "DATA FOR diameter " + randomdiameter + " with " + randompoints + " randompoints with all faces accessible and failcount of " + failthreshold + " with accessibility requirement REMOVED\n\n";
				for (int t=0; t<6; t++)
				{
					d += trackdescrs [t] + ":\n" + tracklevels [t] + "\n\n";
				}
				Debug.Log(d);
				
				randompoints++;
				//randompoints = randomdiameter * randomdiameter * randomdiameter / 6;
				track = new float[]{-1.0f,-1.0f,10000,-1.0f,-1.0f,-1.0f};
				tracklevels = new string[]{"","","","","",""};				
			}
			
			if (Input.GetKeyDown(KeyCode.Y))
			{
				Application.runInBackground = false;
				break;
			}
			FastRandomLevel();
			List<Face> accessiblefaces = AccessibleFaces();
			if (accessiblefaces.Count == Faces().Count)
			{
				int connectedcomponents = ConnectedComponents(accessiblefaces).Count;
				if (connectedcomponents == 1)
				{
					//int connectedpointcomponents = ConnectedPointComponents().Count;
					float deadendchance;
					float averagenumberofdirs;
					int deg8faces;
					float univiewablefacechance;
					int parallelmovements;
					float chancehalfcovered;
					float chancebihalfcovered;
					FaceStatsAggregate(accessiblefaces, out deadendchance, out averagenumberofdirs, out deg8faces, out univiewablefacechance, out parallelmovements, out chancehalfcovered, out chancebihalfcovered);
					
					if (deadendchance > track [0])
					{
						track [0] = deadendchance;
						tracklevels [0] = Serialize();
						failcount = 0;
					}
					
					int vertexconnectedcount = VertexConnectedPointComponents().Count;
					if (vertexconnectedcount > track [1])
					{
						track [1] = vertexconnectedcount;
						tracklevels [1] = Serialize();
						failcount = 0;
					}
					if (averagenumberofdirs < track [2])
					{
						track [2] = averagenumberofdirs;
						tracklevels [2] = Serialize();
						failcount = 0;
					}
					
					if (univiewablefacechance > track [3])
					{
						track [3] = univiewablefacechance;
						tracklevels [3] = Serialize();
						failcount = 0;
					}
					
					if (chancehalfcovered > track [4])
					{
						track [4] = chancehalfcovered;
						tracklevels [4] = Serialize();
						failcount = 0;
					}
					
					if (chancebihalfcovered > track [5])
					{
						track [5] = chancebihalfcovered;
						tracklevels [5] = Serialize();
						failcount = 0;
					}
					failcount++;					
				}
			}
			if (i % 10 == 0)
			{
				i = 0;
				yield return null;
			}
		}
		
		//sw.Stop();
		
		//Debug.Log("time taken " + sw.Elapsed);
		
	}*/

	public void Update ()
	{
		/*
		if (Input.GetKeyDown(KeyCode.R))
		{
			RandomLevel();
		}		
		if (Input.GetKeyDown(KeyCode.T))
		{
			StartCoroutine(GenerateInteresting());
		}		
		if (Input.GetKeyDown(KeyCode.S))
		{
			PrintLevelStats();
		}
		*/
		/*
		if (Input.GetKeyDown(KeyCode.L))
		{
			Deserialize(Editor.blanklevel);
		}
		if (Input.GetKeyDown(KeyCode.M))
		{
			Deserialize("30|-1,-2,-1|-2,0,-1|1,-1,-1|-1,-1,0|1,-2,-1|-3,0,0|0,-2,-1|-1,-2,0|-4,0,0|-2,-1,0|-2,-2,0|-1,0,-1|-2,1,-1|-3,1,0|-4,1,0|-5,0,1|-3,-1,1|-3,-2,1|-5,1,1|-1,1,-1|0,1,-2|0,0,-2|2,-1,-2|2,-2,-2|4,-3,-2|3,-3,-1|2,-3,-1|1,-3,0|0,-3,0|-1,-3,1|0|0,-2,-1^3|4|-0.06200376~0.6740631~-0.7360668");
		}*/
	}
	
	/*
	public int randompoints=30;
	public int randomdiameter=4;
	
	public void PopulateWithRandomPoints ()
	{
		points = new List<Coord>();
		
		int i = 0;
		while (points.Count<randompoints)
		{
			i++;
			if (i > 1000)
			{
				break;
			}
			int dl = -randomdiameter / 2;
			
			Coord c = new Coord(UnityEngine.Random.Range(dl, dl + randomdiameter), UnityEngine.Random.Range(dl, dl + randomdiameter), UnityEngine.Random.Range(dl / 2, (dl + randomdiameter) / 2));

			if (points.IndexOf(c) < 0)
			{
				points.Add(c);
			}
		}		
		//Jitter();
	}
	*/
	
	private void ClearCollectables ()
	{
		foreach (GameObject collectable in collectable_obs)
		{
			Destroy(collectable);
		}
		collectable_obs = new List<GameObject>();
		collectables = new List<Face>();
	}
	
	private void GenerateCollectables (int count)
	{
		for (int i = 0; i < count; i++)
		{
			int rc = UnityEngine.Random.Range(0, dodecas.Count);
			int rf = UnityEngine.Random.Range(0, dodecas[rc].faces.Count);
			Face f = dodecas[rc].faces[rf];
			if (playerface.ValueEquals(f)==false)
			{
				collectables.Add(f);
			}
		}
		SpawnCollectables();
	}
	
	public void SpawnCollectables ()
	{
		foreach (GameObject collectable in collectable_obs)
		{
			Destroy(collectable);
		}
		collectable_obs = new List<GameObject>();
		
		foreach (Face f in collectables)
		{
			Vector3 pos = f.PushedCenter();
			int s = f.side;
			Quaternion rot = Quaternion.LookRotation(DodecaData.SideForward [s], DodecaData.SideNormal [s]);
			GameObject go = Instantiate(collectableprefab, pos, rot) as GameObject;
			go.GetComponent<Collectable>().face = f;
			go.transform.parent = gameObject.transform;
			collectable_obs.Add(go);			
		}
		initcollectablecount = collectables.Count;
		UpdateCollectableCount();
	}
	
	public GameObject collectableprefab;
	
	public void HideInvisibleCollectables ()
	{
		Vector3 camforward = - DodecaData.vertices [mo.targetview];
		foreach (GameObject go in collectable_obs)
		{
			Face f = go.GetComponent<Collectable>().face;
			float angle = Vector3.Angle(DodecaData.SideNormal [f.side], camforward);
			if (angle < 95)
			{
				go.renderer.enabled = false;
			}
		}
	}
	
	public void ShowInvisibleCollectables ()
	{
		foreach (GameObject go in collectable_obs)
		{
			go.renderer.enabled = true;
		}
	}
	
	
	public bool HasFace (Face f)
	{
		foreach (Dodeca d in dodecas)
		{
			foreach (Face f2 in d.faces)
			{
				if (f2.ValueEquals(f))
				{				
					return true;
				}					
			}
		}
		return false;
	}
	
	//doesn't despawn
	private void CullBlockedOffCollectables ()
	{
		for (int i = collectables.Count-1; i>=0; i--)
		{
			Face f = collectables [i];
			if (HasFace(f) == false)
			{
				collectables.RemoveAt(i);
			}
		}
	}

	public void Regen ()
	{
		ClearCollectables();
		dodecas = new List<Dodeca>();
		foreach (Coord p in points)
		{
			dodecas.Add(new Dodeca(p));
		}		
		
		CullEmptyDodecas(dodecas,points);
		CullBlockedFaces(dodecas,points);
		if (HasFace(playerface) == false)
		{
			playerface = dodecas [0].faces [0];
		}
		CullBlockedOffCollectables();
		GenerateMesh();
		GetComponent<Controller>().SnapPlayer();
		if (Editor.editmode == false)
		{
			mo.Target();
			mo.Snap();
		}
	}
	
	public void TryCollect (Face f)
	{
		for (int i=collectable_obs.Count-1; i>=0; i--)
		{
			GameObject go = collectable_obs [i];
			Collectable c = go.GetComponent<Collectable>();
			if (c.face.ValueEquals(f))
			{
				collectable_obs.Remove(go);
				collectables.RemoveAt(i);
				FadeAndDie fad = go.AddComponent<FadeAndDie>();
				fad.level = this;
				bool finished = Finished();
				fad.last = finished;
			}
		}		
	}	
	
	public bool Finished ()
	{
		return (collectables.Count==0);
	}
	
	public GameObject levelcompletetext;
	
	public void UpdateCollectableCount ()
	{
		text.GetComponent<TextMesh>().text = "Level " + (TitleOptions.curlevel+1)+"/"+LevelData.LevelDats.Length+"\nCollected "+(initcollectablecount-collectables.Count)+"/"+initcollectablecount;
	}

	public void GenerateMesh ()
	{
		MeshData md = new MeshData();
		
		foreach (Dodeca d in dodecas)
		{
			foreach (Face f in d.faces)
			{
				md.AddQuad(f.QuadPoints());
			}
		}
		
		MeshFilter mf = GetComponent<MeshFilter>();
		if (mf.sharedMesh == null)
		{
			mf.sharedMesh = new Mesh();
		}
		
		md.BuildSingleMesh(false, true, mf);
		GetComponent<MeshCollider>().sharedMesh = mf.sharedMesh;
	}
	
	/*
	public Face GetFaceFromNormalRay (Vector3 point, Vector3 normal)
	{
		Vector3 orig_approx = point - normal * 1.4f;
		
		float min = 1000;
		Dodeca target = null;
		foreach (Dodeca cand in dodecas)
		{
			float dist = Vector3.Distance(orig_approx, cand.coord.ToVector3());
			if (dist < min)
			{
				min = dist;
				target = cand;
			}
		}
		min = 1000;
		Face targetface = default(Face);
		foreach (Face f in target.faces)
		{
			float a = Vector3.Angle(normal, DodecaData.SideNormal [f.side]);
			if (a < min)
			{
				min = a;
				targetface = f;
			}
		}
		return targetface;
	}
	*/
	
	//more fragile than the above
	public Face GetFaceFromNormalRay (Vector3 point, Vector3 normal)
	{
		Vector3 orig_approx = point - normal * 1.4f;
		
		Coord coord = Coord.FromVector3(orig_approx);
		
		int side = -1;
		float mind = 1000;
		for (int i=0; i<12; i++)
		{
			float d = Vector3.Distance(normal, DodecaData.SideNormal [i]);
			if (d < mind)
			{
				side = i;
				mind = d;
			}
		}
		
		return new Face(coord, side);
	}
	
}
