using UnityEngine;
using System;
using System.Collections.Generic;

public class MeshData
{
	private struct Triple
	{
		public readonly int x;
		public readonly int y;
		public readonly int z;
		public Triple (int _x, int _y, int _z)
		{
			x = _x;
			y = _y;
			z = _z;			
		}
	}
	private class MeshChunk
	{
		public List<Vector3> vertices = new List<Vector3>();
		public List<Vector3> normals = new List<Vector3>();
		public List<Vector2> uvs = new List<Vector2>();
		public List<Triple> triangles = new List<Triple>();
		
		public bool AtCapacity ()
		{
			return vertices.Count > 64500;
		}
		
		public void BuildMesh (bool perlin, bool calctangents, MeshFilter mf)
		{
			//Debug.Log("build mesh");
			//perlin = false;
			Mesh m = new Mesh();
		
			if (perlin)
			{
				
				/*Perlin p = new Perlin();
				for (int i = 0; i < vertices.Count; i++)
				{
					vertices [i] += normals [i] * (p.Noise(0.81f * vertices [i])) * 0.05f;
				}*/
			}
			
			
			m.vertices = vertices.ToArray();
			m.uv = uvs.ToArray();
			
			if (calctangents)
			{
				List<Vector4> tangents = new List<Vector4>();
				for (int i = 0; i < vertices.Count; i++)
				{
					tangents.Add(new Vector4(UnityEngine.Random.Range(0.0f, 0.1f), UnityEngine.Random.Range(0.0f, 0.1f), UnityEngine.Random.Range(0.0f, 0.1f), UnityEngine.Random.Range(0.0f, 0.1f)));
				}
				m.tangents = tangents.ToArray();
			}
			
			if (normals.Count > 0)
			{
				m.normals = normals.ToArray();
			}
			
			int[] trilist = new int[triangles.Count * 3];
			for (int i = 0; i < triangles.Count; i++)
			{
				trilist [3 * i + 0] = triangles [i].x;
				trilist [3 * i + 1] = triangles [i].z;
				trilist [3 * i + 2] = triangles [i].y;
			}
			m.triangles = trilist;
			if (normals.Count == 0)
			{
				m.RecalculateNormals();
			}
			
			//m.Optimize();
			
			//Debug.Log(mf.name);
			//Debug.Log("vertex count : " + m.vertices.Length);
			//Debug.Log(mf.name);
			if (mf.sharedMesh != null)
			{				
				GameObject.Destroy(mf.sharedMesh);
			}
			
			mf.sharedMesh = m;
			//GameObject.Destroy(m);
		}
		
		public Mesh BuildMesh (bool perlin, bool calctangents)
		{
			//perlin = false;
			Mesh m = new Mesh();
			if (perlin)
			{
				/*
				Perlin p = new Perlin();
				for (int i = 0; i < vertices.Count; i++)
				{
					vertices [i] += normals [i] * (p.Noise(0.81f * vertices [i])) * 0.05f;
				}*/
			}
			
			
			m.vertices = vertices.ToArray();
			m.uv = uvs.ToArray();
			
			if (calctangents)
			{
				List<Vector4> tangents = new List<Vector4>();
				for (int i = 0; i < vertices.Count; i++)
				{
					tangents.Add(new Vector4(UnityEngine.Random.Range(0.0f, 0.1f), UnityEngine.Random.Range(0.0f, 0.1f), UnityEngine.Random.Range(0.0f, 0.1f), UnityEngine.Random.Range(0.0f, 0.1f)));
				}
				m.tangents = tangents.ToArray();
			}
			
			if (normals.Count > 0)
			{
				m.normals = normals.ToArray();
			}
			
			int[] trilist = new int[triangles.Count * 3];
			for (int i = 0; i < triangles.Count; i++)
			{
				trilist [3 * i + 0] = triangles [i].x;
				trilist [3 * i + 1] = triangles [i].z;
				trilist [3 * i + 2] = triangles [i].y;
			}
			m.triangles = trilist;
			if (normals.Count == 0)
			{
				m.RecalculateNormals();
			}
			//Debug.Log("vertex count : " + m.vertices.Length);
			return m;
		}
	}
	
	List<MeshChunk> chunks;
	MeshChunk currentchunk;
		
	public MeshData ()
	{
		chunks = new List<MeshChunk>();
		currentchunk = new MeshChunk();
		chunks.Add(currentchunk);
	}
	
	private void TryRebatch ()
	{
		if (currentchunk.AtCapacity())
		{
			currentchunk = new MeshChunk();
			chunks.Add(currentchunk);
		}
	}
	
	private Vector3 Centroid ()
	{
		Vector3 result = new Vector3(0, 0, 0);
		int count = 0;
		foreach (MeshChunk c in chunks)
		{
			foreach (Vector3 v in c.vertices)
			{
				result += v;
				count++;
			}
		}
		result /= count;
		return result;
	}
	
	/*
	//fits within sphere of radius and centers at 0,0,0
	public void Normalize (float radius, out Vector3 centroidoffset, out float radiusmultiplier)
	{
		Vector3 centroid = Centroid();
		float radiusmax = 0.0f;
		int i;
		
		foreach (MeshChunk chunk in chunks)
		{						
			for (i=0; i < chunk.vertices.Count; i++)
			{
				chunk.vertices [i] -= centroid;
				float d = chunk.vertices [i].magnitude;
				if (d > radiusmax)
				{
					radiusmax = d;
				}
			}
		}
		
		foreach (MeshChunk chunk in chunks)
		{				
			for (i=0; i < chunk.vertices.Count; i++)
			{
				chunk.vertices [i] *= radius / radiusmax;
			}
		}
		
		centroidoffset = centroid;
		radiusmultiplier = radius / radiusmax;
	}*/
	
	public int TriangleCount ()
	{
		int sum = 0;
		foreach (MeshChunk chunk in chunks)
		{
			sum += chunk.triangles.Count;
		}
		return sum;
	}
	
	//assumes toAppend has only one chunk
	public void Append (MeshData toAppend)
	{
		MeshChunk appendchunk = toAppend.currentchunk;
		
		int oldlen = currentchunk.vertices.Count;
		currentchunk.vertices.AddRange(appendchunk.vertices);
		currentchunk.uvs.AddRange(appendchunk.uvs);
		currentchunk.normals.AddRange(appendchunk.normals);
		
		for (int i = 0; i < appendchunk.triangles.Count; i++)
		{
			Triple tri = appendchunk.triangles [i];
			currentchunk.triangles.Add(new Triple(tri.x + oldlen, tri.y + oldlen, tri.z + oldlen));
		}
		
		TryRebatch();		
	}
	
	private void AddVertex (Vector3 v, Vector2 uv)
	{
		currentchunk.vertices.Add(v);
		currentchunk.uvs.Add(uv);
	}
	
	private void AddVertex (Vector3 v, Vector2 uv, Vector3 n)
	{
		currentchunk.vertices.Add(v);
		currentchunk.uvs.Add(uv);
		currentchunk.normals.Add(n.normalized);
	}
	
	private void AddVertex (Vector3 v)
	{
		currentchunk.vertices.Add(v);
		currentchunk.uvs.Add(new Vector2(0, 0));
	}
	
	private void AddTriangle (Triple t)
	{
		currentchunk.triangles.Add(t);
	}
	
	/*
	 * public void AddTriangle (Vector3 a, Vector3 b, Vector3 c)
	{
		AddTriangle(new Triple(vertices.Count, vertices.Count + 1, vertices.Count + 2));
		AddVertex(c);
		AddVertex(b);
		AddVertex(a);
	}*/
		
	public void AddNormaledTri (Vector3 a, Vector3 b, Vector3 c, Vector3 na, Vector3 nb, Vector3 nc)
	{
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 1, currentchunk.vertices.Count + 2));
		AddVertex(a, new Vector2(0, 1), na);
		AddVertex(b, new Vector2(1, 1), nb);
		AddVertex(c, new Vector2(1, 0), nc);
		
		
		TryRebatch();
	}
	
	public void AddNormaledQuad (Vector3 a, Vector3 b, Vector3 c, Vector3 d, Vector3 na, Vector3 nb, Vector3 nc, Vector3 nd, Vector2 uva, Vector2 uvb,
	Vector2 uvc, Vector2 uvd)
	{
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 1, currentchunk.vertices.Count + 2));
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 2, currentchunk.vertices.Count + 3));
		AddVertex(a, uva, na);
		AddVertex(b, uvb, nb);
		AddVertex(c, uvc, nc);
		AddVertex(d, uvd, nd);
		
		
		TryRebatch();
	}
	
	public void AddQuad (Vector3[] dat)
	{
		Vector2 uva = new Vector2(dat [4].x, dat [4].y);
		Vector2 uvb = new Vector2(dat[5].x,dat[5].y);
		Vector2 uvc = new Vector2(dat[6].x,dat[6].y);
		Vector2 uvd = new Vector2(dat[7].x,dat[7].y);
		AddQuad(dat[3],dat[1],dat[2],dat[0],uvd,uvb,uvc,uva);
	}
	
	public void AddQuad (Vector3 a, Vector3 b, Vector3 c, Vector3 d)
	{
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 1, currentchunk.vertices.Count + 2));
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 2, currentchunk.vertices.Count + 3));
		AddVertex(a, new Vector2(0, 1));
		AddVertex(b, new Vector2(1, 1));
		AddVertex(c, new Vector2(1, 0));
		AddVertex(d, new Vector2(0, 0));
		
		
		TryRebatch();
	}
	
	public void AddCube (Vector3 center, float width)
	{
		float w = width / 2;
		//top face
		AddQuad(center + new Vector3(w, w, w), center + new Vector3(-w, w, w), center + new Vector3(-w, w, -w), center + new Vector3(w, w, -w));
		
		//bottom face
		AddQuad(center + new Vector3(w, -w, -w), center + new Vector3(-w, -w, -w), center + new Vector3(-w, -w, w), center + new Vector3(w, -w, w));
				
		//left face
		AddQuad(center + new Vector3(-w, -w, w), center + new Vector3(-w, -w, -w), center + new Vector3(-w, w, -w), center + new Vector3(-w, w, w));
				
		//right face
		AddQuad(center + new Vector3(w, w, w), center + new Vector3(w, w, -w), center + new Vector3(w, -w, -w), center + new Vector3(w, -w, w));
							
		//forward face
		AddQuad(
			center + new Vector3(w, w, w), 
			center + new Vector3(w, -w, w), 
			center + new Vector3(-w, -w, w), 
			center + new Vector3(-w, w, w)
			);

		//back face
		AddQuad(
			center + new Vector3(w, -w, -w), 
			center + new Vector3(w, w, -w), 
			center + new Vector3(-w, w, -w), 
			center + new Vector3(-w, -w, -w)
			);
		
		
		TryRebatch();
	}
	
	[Flags]
	public enum CubeSide
	{
		None = 0,
		Front = 1,
		Back = 2,
		Left = 4,
		Right = 8,
		Top = 16,
		Bottom = 32,
	}
	
	public void AddCube (Vector3 center, float rx, float ry, float rz, Vector2 uv, CubeSide sidehide)
	{
		//top face
		if ((sidehide & CubeSide.Top) == CubeSide.None)
		{
			AddQuad(center + new Vector3(rx, ry, rz), center + new Vector3(-rx, ry, rz), center + new Vector3(-rx, ry, -rz), center + new Vector3(rx, ry, -rz), uv, uv, uv, uv);
		}
		
		//bottom face
		if ((sidehide & CubeSide.Bottom) == CubeSide.None)
		{
			AddQuad(center + new Vector3(rx, -ry, -rz), center + new Vector3(-rx, -ry, -rz), center + new Vector3(-rx, -ry, rz), center + new Vector3(rx, -ry, rz), uv, uv, uv, uv);
		}
		
		//left face
		if ((sidehide & CubeSide.Left) == CubeSide.None)
		{
			AddQuad(center + new Vector3(-rx, -ry, rz), center + new Vector3(-rx, -ry, -rz), center + new Vector3(-rx, ry, -rz), center + new Vector3(-rx, ry, rz), uv, uv, uv, uv);
		}
		
		//right face
		if ((sidehide & CubeSide.Right) == CubeSide.None)
		{
			AddQuad(center + new Vector3(rx, ry, rz), center + new Vector3(rx, ry, -rz), center + new Vector3(rx, -ry, -rz), center + new Vector3(rx, -ry, rz), uv, uv, uv, uv);
		}
		
		//forward face
		if ((sidehide & CubeSide.Back) == CubeSide.None)
		{
			AddQuad(center + new Vector3(rx, ry, rz), center + new Vector3(rx, -ry, rz), center + new Vector3(-rx, -ry, rz), center + new Vector3(-rx, ry, rz), uv, uv, uv, uv);
		}
		
		//back face
		if ((sidehide & CubeSide.Front) == CubeSide.None)
		{
			AddQuad(center + new Vector3(rx, -ry, -rz), center + new Vector3(rx, ry, -rz), center + new Vector3(-rx, ry, -rz), center + new Vector3(-rx, -ry, -rz), uv, uv, uv, uv);
		}
		
		
		TryRebatch();
	}
	
	public void AddQuads (Vector3[,] newvertices, Vector2[,] uvs)
	{

		for (int i = 0; i < newvertices.GetLength(0); i++)
		{
			for (int j = 0; j < newvertices.GetLength(1); j++)
			{
				AddVertex(newvertices [i, j], uvs [i % 2, j % 2]);
				
				//starting at top-left corner
				if (i < newvertices.GetLength(0) - 1 && j < newvertices.GetLength(1) - 1)
				{
					//Debug.Log("adding quad" + i + "," + j);
					int p1 = currentchunk.vertices.Count - 1;
					int p2 = currentchunk.vertices.Count - 1 + 1;
					int p3 = currentchunk.vertices.Count - 1 + 1 + newvertices.GetLength(1);
					int p4 = currentchunk.vertices.Count - 1 + newvertices.GetLength(1);
					AddTriangle(new Triple(p1, p2, p3));
					AddTriangle(new Triple(p1, p3, p4));
				}
			}
		}
		
		TryRebatch();
		/*	
		AddTriangle(new Triple(vertices.Count, vertices.Count + 1, vertices.Count + 2));
		AddTriangle(new Triple(vertices.Count, vertices.Count + 2, vertices.Count + 3));
		AddVertex(a, new Vector2(0, 1));
		AddVertex(b, new Vector2(1, 1));
		AddVertex(c, new Vector2(1, 0));
		AddVertex(d, new Vector2(0, 0));*/
	}
	
	public void AddQuad (Vector3 a, Vector3 b, Vector3 c, Vector3 d, Vector2 uva, Vector2 uvb, Vector2 uvc, Vector2 uvd)
	{
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 1, currentchunk.vertices.Count + 2));
		AddTriangle(new Triple(currentchunk.vertices.Count, currentchunk.vertices.Count + 2, currentchunk.vertices.Count + 3));
		AddVertex(a, uva);
		AddVertex(b, uvb);
		AddVertex(c, uvc);
		AddVertex(d, uvd);
		
		TryRebatch();
	}
		
	public void BuildSingleMesh (bool perlin, bool calctangents, MeshFilter mf)
	{			
		chunks [0].BuildMesh(perlin, calctangents, mf);
	}
	
	//calctangents = whether to generate random tangents
	public Mesh BuildSingleMesh_Dangerous (bool perlin, bool calctangents)
	{	
		return chunks [0].BuildMesh(perlin, calctangents);
	}
	
	public List<Mesh> BuildMesh (bool perlin, bool calctangents)
	{
		
		List<Mesh> result = new List<Mesh>();
		foreach (MeshChunk chunk in chunks)
		{
			result.Add(chunk.BuildMesh(perlin, calctangents));
		}
		return result;
	}

	public static void AddTangents (Mesh mesh)
	{
		Vector3[] tan2 = new Vector3[mesh.vertices.Length];
		Vector3[] tan1 = new Vector3[mesh.vertices.Length];
		Vector4[] tangents = new Vector4[mesh.vertices.Length];
		//Vector3[] binormal = new Vector3[mesh.vertices.Length];
		for (int a = 0; a < (mesh.triangles.Length); a += 3)
		{
			long i1 = mesh.triangles [a + 0];
			long i2 = mesh.triangles [a + 1];
			long i3 = mesh.triangles [a + 2];
	
			Vector3 v1 = mesh.vertices [i1];
			Vector3 v2 = mesh.vertices [i2];
			Vector3 v3 = mesh.vertices [i3];
	
			Vector2 w1 = mesh.uv [i1];
			Vector2 w2 = mesh.uv [i2];
			Vector2 w3 = mesh.uv [i3];
	
			float x1 = v2.x - v1.x;
			float x2 = v3.x - v1.x;
			float y1 = v2.y - v1.y;
			float y2 = v3.y - v1.y;
			float z1 = v2.z - v1.z;
			float z2 = v3.z - v1.z;
	
			float s1 = w2.x - w1.x;
			float s2 = w3.x - w1.x;
			float t1 = w2.y - w1.y;
			float t2 = w3.y - w1.y;
	
			float r = 1.0F / (s1 * t2 - s2 * t1);
			Vector3 sdir = new Vector3((t2 * x1 - t1 * x2) * r, (t2 * y1 - t1 * y2) * r, (t2 * z1 - t1 * z2) * r);
			Vector3 tdir = new Vector3((s1 * x2 - s2 * x1) * r, (s1 * y2 - s2 * y1) * r, (s1 * z2 - s2 * z1) * r);
	
			tan1 [i1] += sdir;
			tan1 [i2] += sdir;
			tan1 [i3] += sdir;
	
			tan2 [i1] += tdir;
			tan2 [i2] += tdir;
			tan2 [i3] += tdir;
		}
	
		for (int a = 0; a < mesh.vertices.Length; a++)
		{
			Vector3 n = mesh.normals [a];
			Vector3 t = tan1 [a];
	
			Vector3.OrthoNormalize(ref n, ref t);
			tangents [a].x = t.x;
			tangents [a].y = t.y;
			tangents [a].z = t.z;
	
			// Calculate handedness
			tangents [a].w = (Vector3.Dot(Vector3.Cross(n, t), tan2 [a]) < 0.0f) ? -1.0f : 1.0f;
	
			//To calculate binormals if required as vector3 try one of below:-
			//Vector3 binormal[a] = (Vector3.Cross(n, t) * tangents[a].w).normalized;
			//Vector3 binormal[a] = Vector3.Normalize(Vector3.Cross(n, t) * tangents[a].w)
		}
		mesh.tangents = tangents;
	}	
}
