using UnityEngine;
using System.Collections.Generic;
using System.Collections;
using System;

class MouseOrbit : MonoBehaviour
{
	public Level level = null;
	private Controller controller;
	public Transform target = null;
	public float distance = 10.0f;
	//public float distanceMin = 5.0f;
	//public float distanceMax = 100.0f;
	public float wheelsensitivity = 10.0f;
	public float xSpeed = 250.0f;
	public float ySpeed = 120.0f;
	public Camera cam;
	
	void Awake ()
	{
		controller = level.GetComponent<Controller>();
		cam = GetComponent<Camera>();
		position = transform.position;
		rotation = transform.rotation;	
		
	}
	
	void Start ()
	{
	
		// Make the rigid body not change rotation
		if (rigidbody)
		{
			rigidbody.freezeRotation = true;
		}
	}
		
	private Vector2  lastm = Vector2.zero;
		
	private Vector2 GetInput ()
	{
		float dx = 0;
		float dy = 0;
		
		if (target != null)
		{
			Vector2 m = InputManager.GetAxis();
			
#if !UNITY_IPHONE
			if (Mathf.Abs(m.x) > 0.35 || Mathf.Abs(m.y) >= 0.35)
			{
				m = lastm;
				lastm = Vector2.zero;
			}
			else
			{
				lastm = m;
			}			
#endif
			dx = m.x * xSpeed;
			dy = m.y * ySpeed;
			
		}
		Vector2 result = new Vector2(dx, dy);
		
		return result;
	}
	
	public Vector3 DesiredPos ()
	{
		//distance -= Input.GetAxis("Mouse ScrollWheel") * Time.deltaTime * wheelsensitivity;
		//distance = Mathf.Clamp(distance, distanceMin, distanceMax);
		
		Vector2 d = GetInput();
		
		rotation = transform.rotation * Quaternion.Euler(-d.y, d.x, 0);
		position = rotation * new Vector3(0.0f, 0.0f, -distance) + target.position;
		
		return position;
	}
	
	private bool PlayerVisible (int i)
	{
		Vector3 cand = DodecaData.vertices [i];
		Vector3 candforward = -cand;
		Face f = level.playerface;
		Vector3 n = DodecaData.SideNormal[f.side];
		float angle = Vector3.Angle(cand, n);
		if (angle > 85)
		{
			return false;
		}
		//now raycast to check it's actually visible (it = at least one of the half-way vectors)
		
		Vector3[] edgedirs = DodecaData.EdgeDirs[(int)f.side];		
		foreach (Vector3 v in edgedirs)
		{
			//half way must be visible - i.e., must give face f
			Vector3 halfway = f.Center() + v * 1.1f / 2.0f;
			Ray halfwayray = new Ray(halfway - 100 * candforward, candforward);
			RaycastHit hit;
			if (Physics.Raycast(halfwayray, out hit, 1000, 1 << 10))
			{
				Face newface = level.GetFaceFromNormalRay(hit.point, hit.normal);
				bool result = level.playerface.ValueEquals(newface);
				if (result == true)
				{
					return true;
				}
				else
				{
				}						
			}
		}
		
		return false;		
	}
	
	private int GetClosestDirIndex (out Vector3 dir, out Vector3 up, out int upindex)
	{
		//get the nearest "forward" vector.
		up = default(Vector3);
		Vector3 forward = transform.forward;
		int target = 0;
		Vector3 targetvec = targetforward;
		float min = 1000;
		for (int i2 = 0; i2 < 8; i2++)
		{
			int i = DodecaData.validviewdirs [i2];
			
			Vector3 cand = DodecaData.vertices [i];
			float angle = Vector3.Angle(cand, -forward);


			//filters out bad points(that give checkered grids)...lazy but it works!
			if (angle < min)
			{
				if (PlayerVisible(i))
				{
					targetvec = -cand;
					min = angle;
					target = i;
				}	
			}
		}	
		
		//get the nearest "up" vector (not currently used)
		float minangle = 100;
		int minangleindex = -1;
		//iterate through all directions
		for (int i=0; i<12; i++)
		{
			//if this directions has an appropriate triangle
			for (int j=0; j<4; j++)
			{
				if (DodecaData.dodec_tris [i * 4 + j] == target)
				{					
					Vector3 cand_dir = DodecaData.SideNormal[i];
					float a = Vector3.Angle(cand_dir, transform.up);
					if (a < minangle)
					{
						up = cand_dir;
						minangle = a;
						minangleindex = i;
					}					
				}
			}		
		}
		upindex = minangleindex;
		
		dir = targetvec;
		
		return target;
	}
	
	public float time;
	public bool tweening=false;
	private Vector3 startforward;
	public Vector3 targetforward;
	public int targetview;
	
	public int targetupindex;
	public Vector3 targetup;
	public Vector3 startup;
	
	public enum CameraState {
		iso,
		rotating,
		tweening
	}
	public CameraState state = CameraState.tweening;
		
	public float tweenlength = 3f;
		
	public void Target ()
	{
		targetview = GetClosestDirIndex(out targetforward, out targetup, out targetupindex);
		state = CameraState.tweening;
		startforward = transform.forward;
		startup = transform.up;
		time = tweenlength;
	}
	
	public void SnapWithUp ()
	{
		time = -1;
		Vector3 cand;
		if (targetview < 0 || targetview >= 14)
		{
			Debug.Log("eek");
			cand=Vector3.up;
		}
		else
		{
			cand = DodecaData.vertices [targetview];
		}
		targetforward = -cand;
		rotation = Quaternion.LookRotation(targetforward, targetup);
		state = CameraState.iso;
		level.HideInvisibleCollectables();			
	}
	
	public void Snap ()
	{
		time = -1;
		Vector3 cand;
		if (targetview < 0 || targetview >= 14)
		{
			Debug.Log("eek");
			cand = Vector3.up;
		}
		else
		{
			cand = DodecaData.vertices [targetview];
		}
		targetforward = -cand;
		rotation = Quaternion.LookRotation(targetforward, transform.up);
		level.HideInvisibleCollectables();	
	}
	
	/*List<Vector3> invalidviews = new List<Vector3>();
	
	private void CalcInvalidViews ()
	{
		invalidviews = new List<Vector3>();
		for (int i=0; i<14; i++)
		{
			if (PlayerVisible(i) == false)
			{
				invalidviews.Add(-DodecaData.vertices[i]);
			}
		}
	}*/
	
	public void CalcCoords ()
	{
		//		distance -= Input.GetAxis("Mouse ScrollWheel") * Time.deltaTime * wheelsensitivity;
		//distance = Mathf.Clamp(distance, distanceMin, distanceMax);		
		
		if (InputManager.JustPressedRotate() && controller.animdone)
		{
			//CalcInvalidViews();
			state = CameraState.rotating;
			level.ShowInvisibleCollectables();
		}
		
				
		if (state == MouseOrbit.CameraState.rotating && InputManager.JustReleasedRotate())
		{
			if (Editor.editmode)
			{
				state = CameraState.iso;		
			}
			else
			{
				Target();
			}
		}
		
		if (state == CameraState.tweening)
		{
			time -= Time.deltaTime;
			if (time < 0)
			{
				state = CameraState.iso;
				Vector3 forward = targetforward;
				rotation = Quaternion.LookRotation(forward, transform.up);
				level.HideInvisibleCollectables();			
			}
			else
			{
				float t = time / tweenlength;
				Vector3 forward = Vector3.Slerp(targetforward, startforward, t);
				Vector3 up = transform.up;
				// Vector3.Slerp(targetup, startup, t);
				rotation = Quaternion.LookRotation(forward, up);
			}
			//public float t = Time.cu tweenstart
			
			//	RefreshPlayer();
		}
		
		if (state == CameraState.rotating)
		{
			Vector2 d = GetInput();
			rotation = transform.rotation * Quaternion.Euler(-d.y, d.x, 0);
			
			//	RefreshPlayer();
		}
		
		position = rotation * new Vector3(0.0f, 0.0f, -distance) + target.position;
		

		transform.rotation = rotation;		
		transform.position = position;
	
	}
	
	/*
	private void RefreshPlayer ()
	{
		float mina = 1000;
		Vector3 forward = transform.forward;
		foreach (Vector3 invalid in invalidviews)
		{
			float angle = Vector3.Angle(forward, invalid);
			if (angle < mina)
			{
				mina = angle;
			}
		}
		if (mina < 40)
		{
			float b = mina / 40;
			controller.player.renderer.material.color = new Color(b,b,b,1) ;
		}
		else
		{
			controller.player.renderer.material.color = Color.white;
		}
		Debug.Log(mina);
	}
	*/
	[HideInInspector]
	public Quaternion rotation;
	[HideInInspector]
	public Vector3 position;
	
	void LateUpdate ()
	{	
		if (level.leaving)
		{
			return;
		}
		CalcCoords();		    
	}
}
