using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class Player : MonoBehaviour {
	
	public Material staticmaterial;
	public Material dynamicmaterial;
	
	private MeshFilter mf;
	
	void Awake ()
	{		
		mf = GetComponent<MeshFilter>();
		mf.sharedMesh = new Mesh();
		GenStaticMesh();
	}
	
	public const float contractamount = 0.693f;
	
	private static Vector3[] vertices = {
			new Vector3(-2 * 0.713f, 0, 0) * contractamount,
			new Vector3(0, 0, 1) * contractamount,
			new Vector3(2 * 0.713f, 0, 0) * contractamount,
			new Vector3(0, 0, -1) * contractamount,
		};
	
	public void GenDynamicMesh (Vector3 pivot, out Vector4 pivot_x, out Vector4 pivot_y)
	{		
		List<Vector3> static_v = new List<Vector3>();
		List<Vector3> dynamic_v = new List<Vector3>();
		foreach (Vector3 v in vertices)
		{
			float d = Vector3.Distance(pivot, transform.TransformPoint(v));
			if (d < 1)
			{
				static_v.Add(v);
			}
			else
			{
				dynamic_v.Add(v);
			}
		}		
		
		//1.20699 seems to be "parallel" distance, with next being 1.30something, so make 1.25 the cutoff
		//1.97 is "too large" distance, so use that to decide
		
		{
			float d = Vector3.Distance(
				transform.TransformPoint(static_v [0]), 
				transform.TransformPoint(dynamic_v [0]));
			if (d > 1.5)
			{
				//swap
				Vector3 t = dynamic_v [0];
				dynamic_v [0] = dynamic_v [1];
				dynamic_v [1] = t;				
			}			
		}
		
		
		
		//now they're matched up
		
		Camera cam = Camera.mainCamera;
		
		Matrix4x4 M = transform.localToWorldMatrix;
		Matrix4x4 V = cam.worldToCameraMatrix;
		//Matrix4x4 P = cam.projectionMatrix;		
		
		Matrix4x4 MV = V * M;
		
		Vector3 va = MV * static_v [0];
		Vector3 vb = MV * dynamic_v [0];
		Vector3 diff = vb - va;
		Vector4 diff4 = (new Vector4(diff.x, diff.y, diff.z));
		
		// need to offset by pivot amount
		Vector3 pivot_offset_world = transform.position - pivot;
		Vector3 pivot_offset_model = transform.InverseTransformDirection(pivot_offset_world);
		
		
		Vector3 pivot_model = transform.InverseTransformPoint(pivot);
		Vector3 average_model = (static_v [0] + static_v [1]) / 2;
		Vector3 pivotdiff_model = average_model - pivot_model;
		
		
		
		MeshData md = new MeshData();
		md.AddQuad(
			static_v [0] + pivot_offset_model - pivotdiff_model, 
			static_v [1] + pivot_offset_model - pivotdiff_model, 
			static_v [1] + pivot_offset_model - pivotdiff_model, 
			static_v [0] + pivot_offset_model - pivotdiff_model,
			Vector2.right, Vector2.right, Vector2.up, Vector2.up
			);
		
		//Debug.DrawLine(transform.TransformPoint(average_model), transform.TransformPoint(pivot_model), Color.magenta, 5);
		
		md.BuildSingleMesh(false, true, mf);
		
		//need to move pivotdiff_model to world space
		
		Vector4 pivotdiff_projection = MV.MultiplyVector(pivotdiff_model);
		pivot_x = pivotdiff_projection;
		pivot_y = pivotdiff_projection + diff4;
		
//		pivot_x = new Vector4(pivot_x.x, pivot_x.x, - 0.04f, pivot_x.w);
		pivot_y = new Vector4(pivot_y.x, pivot_y.y,  0.5f, pivot_y.w);
		
		renderer.sharedMaterial = dynamicmaterial;
		renderer.sharedMaterial.SetVector("_ByX", pivot_x);
		renderer.sharedMaterial.SetVector("_ByY", pivot_y);
		
	}
	
	public void GenStaticMesh ()
	{
		if (mf == null)
		{			
			mf = GetComponent<MeshFilter>();
		}
		MeshData md = new MeshData();
		md.AddQuad(vertices [0], vertices [1], vertices [2], vertices [3]);
		
		md.BuildSingleMesh(false, true, mf);
		renderer.sharedMaterial = staticmaterial;
	}
}
